@@ -1,5 +1,6 @@
 import collections
 import contextlib
+import filecmp
 import os
 import re
 import shutil
@@ -17,41 +18,90 @@ __version__ = version.__version__
 RE_PATCH_IDENTIFIER = re.compile('Ply-Patch: (.*)')
 
 
+FROM_SHA1_VALUE = 'ply'
+PATCH_GIT_VERSION = '1.8.3'
+
+
+def _replace_from_sha1(lines):
+    """The SHA1 on the 'From' line of the patch will differ each time the
+    patch file is regenerated. To keep this from causing chatty diffs,
+    replace the SHA1 with a hardcoded value.
+    """
+    for line_idx, line in enumerate(lines):
+        if line.startswith('From'):
+            break
+    else:
+        raise Exception("Malformed patch: 'From' not found")
+
+    parts = lines[line_idx].split(' ')
+    parts[1] = FROM_SHA1_VALUE
+
+    lines[line_idx] = ' '.join(parts)
+
+
+def _replace_git_version(lines):
+    """The version of git is embedded in the patch which will differ causing
+    chatty-diffs and unecessary conflicts.
+
+    It's terribly evil, but hardcoding this is the easiest way to avoid these
+    conflicts.
+    """
+    for reverse_line_idx, line in enumerate(reversed(lines)):
+        if line[0].isdigit() and '.' in line:
+            break
+    else:
+        raise Exception("Malformed patch: Git version not found")
+
+    # The last line counting backwards (0) becomes -(0) - 1 which becomes -1
+    # counting foward
+    line_idx = -reverse_line_idx - 1
+    lines[line_idx] = '%s\n' % PATCH_GIT_VERSION
+
+
+def _remove_ply_patch_annotation(lines):
+    match_idxs = []
+    for idx, line in enumerate(lines):
+        if 'Ply-Patch:' in line:
+            match_idxs.append(idx)
+
+    for idx in reversed(match_idxs):
+        del lines[idx]
+
+
+def _fixup_patch(from_file, to_file):
+    lines = from_file.readlines()
+    _replace_from_sha1(lines)
+    _replace_git_version(lines)
+    _remove_ply_patch_annotation(lines)
+    to_file.write(''.join(lines))
+
+
 class WorkingRepo(git.Repo):
     """Represents our local fork of the upstream repository.
 
     This is where we will create new patches (save) or apply previous patches
     to create a new patch-branch (restore).
     """
+    fetch_remotes = True
+
     def _add_patch_annotation(self, patch_name):
         """Add a patch annotation to the last commit."""
-        self.notes('append', message='Ply-Patch: %s' % patch_name)
+        commit_msg = self.log(count=1, pretty='%B')
+        if 'Ply-Patch' not in commit_msg:
+            commit_msg += '\n\nPly-Patch: %s' % patch_name
+            self.commit(commit_msg, amend=True)
 
-    def _get_patch_annotation(self, notes):
-        """Return the Ply-Patch annotation if present in the git notes.
+    def _get_patch_annotation(self, description):
+        """Return the Ply-Patch annotation if present.
 
         Returns None if not present.
         """
-        matches = re.search(RE_PATCH_IDENTIFIER, notes)
+        matches = re.search(RE_PATCH_IDENTIFIER, description)
         if not matches:
             return None
 
         return matches.group(1)
 
-    def _walk_notes_backwards(self):
-        skip = 0
-        while True:
-            value = self.log(
-                count=1, pretty='%H %N', skip=skip).split(' ', 1)
-
-            if not value[0]:
-                break
-
-            commit_hash, notes = value
-            yield commit_hash, notes
-
-            skip += 1
-
     def _last_upstream_commit_hash(self):
         """Return the hash for the last upstream commit in the repo.
 
@@ -69,6 +119,18 @@ class WorkingRepo(git.Repo):
         return self.log(cmd_arg='%s^' % last_applied_hash,
                         count=1, pretty='%H').strip()
 
+    def _get_commit_hash_and_patch_name(self, cmd_arg=None, count=1,
+                                        skip=None):
+        value = self.log(cmd_arg, count=count, pretty='%H %B',
+                         skip=skip).split(' ', 1)
+
+        if not value[0]:
+            return None, None
+
+        commit_hash, description = value
+        patch_name = self._get_patch_annotation(description)
+        return commit_hash, patch_name
+
     def _applied_patches(self, new_upper_bound=50):
         """Return a list of patches that have already been applied to this
         branch.
@@ -88,8 +150,15 @@ class WorkingRepo(git.Repo):
         """
         applied = []
 
-        for commit_hash, notes in self._walk_notes_backwards():
-            patch_name = self._get_patch_annotation(notes)
+        skip = 0
+        while True:
+            commit_hash, patch_name = self._get_commit_hash_and_patch_name(
+                None, skip=skip)
+
+            skip += 1
+
+            if not commit_hash:
+                break
 
             if patch_name:
                 # Patch found, must be within A
@@ -109,18 +178,9 @@ class WorkingRepo(git.Repo):
     def _store_patch_files(self, patch_names, filenames,
                            parent_patch_name=None):
         """Store a set of patch files in the patch-repo."""
-        for patch_name, filename in zip(patch_names, filenames):
-            # Ensure destination exists (in case a prefix was supplied)
-            dirname = os.path.dirname(patch_name)
-            dest_path = os.path.join(self.patch_repo.path, dirname)
-            if dirname and not os.path.exists(dest_path):
-                os.makedirs(dest_path)
-
-            shutil.move(os.path.join(self.path, filename),
-                        os.path.join(self.patch_repo.path, patch_name))
-
+        source_paths = [os.path.join(self.path, f) for f in filenames]
         return self.patch_repo.add_patches(
-            patch_names, parent_patch_name=parent_patch_name)
+            patch_names, source_paths, parent_patch_name=parent_patch_name)
 
     def _commit_to_patch_repo(self, commit_msg):
         if not self.patch_repo.uncommitted_changes():
@@ -229,8 +289,9 @@ class WorkingRepo(git.Repo):
 
         filenames, parent_patch_name = self._create_patches('HEAD^')
 
-        self._store_patch_files([patch_name], filenames,
-                                parent_patch_name=parent_patch_name)
+        source_paths = [os.path.join(self.path, f) for f in filenames]
+        self.patch_repo.add_patches(
+            [patch_name], source_paths, parent_patch_name=parent_patch_name)
 
         self._add_patch_annotation(patch_name)
         self.restore()  # Apply remaining patches
@@ -283,10 +344,17 @@ class WorkingRepo(git.Repo):
         if self.uncommitted_changes():
             raise exc.UncommittedChanges
 
+        # FIXME: this should probably not be done in the reentrant section and
+        # instead should be done one time, on initializing the restore
+        if self.fetch_remotes:
+            self.fetch(all=True)
+
         applied = set(pn for _, pn in self._applied_patches())
         series = self.patch_repo.series
 
-        for idx, patch_name in enumerate(series):
+        total_applied = len(applied)
+
+        for patch_name in series:
             if patch_name in applied:
                 continue
 
@@ -317,7 +385,9 @@ class WorkingRepo(git.Repo):
             else:
                 self._add_patch_annotation(patch_name)
 
-            sys.stdout.write('\rRestoring %d/%d' % (len(applied) + idx + 1,
+            total_applied += 1
+
+            sys.stdout.write('\rRestoring %d/%d' % (total_applied,
                                                     len(series)))
             sys.stdout.flush()
 
@@ -378,16 +448,12 @@ class WorkingRepo(git.Repo):
             from_filename = os.path.join(self.path, filename)
             with tempfile.NamedTemporaryFile(delete=False) as to_file:
                 with open(from_filename) as from_file:
-                    # len("From fee0d7191da38033ffb29e1d6d88892862064943") =
-                    # 45
-                    first_line = "From ply%s" % from_file.readline()[45:]
-                    to_file.write(first_line)
-                    shutil.copyfileobj(from_file, to_file)
+                    _fixup_patch(from_file, to_file)
 
             shutil.move(to_file.name, from_filename)
 
-        notes = self.log(since, pretty='%N', count=1)
-        parent_patch_name = self._get_patch_annotation(notes)
+        parent_patch_name = self._get_commit_hash_and_patch_name(
+            since)[1]
         return filenames, parent_patch_name
 
     def save(self, since=None, prefix=None):
@@ -424,8 +490,9 @@ class WorkingRepo(git.Repo):
 
             patch_names.append(patch_name)
 
-        added, updated = self._store_patch_files(
-            patch_names, filenames, parent_patch_name=parent_patch_name)
+        source_paths = [os.path.join(self.path, f) for f in filenames]
+        added, updated = self.patch_repo.add_patches(
+            patch_names, source_paths, parent_patch_name=parent_patch_name)
 
         # Remove vestigial patches (anything in the series file after the last
         # recognized patch name)
@@ -517,7 +584,7 @@ class PatchRepo(git.Repo):
 
         self.add('series')
 
-    def add_patches(self, patch_names, parent_patch_name=None):
+    def add_patches(self, patch_names, source_paths, parent_patch_name=None):
         """Add patches to the patch-repo, including add them to the series
         file in the appropriate location.
 
@@ -530,6 +597,28 @@ class PatchRepo(git.Repo):
         """
         added = set()
         updated = set()
+
+        # Move patches into patch-repo
+        for patch_name, source_path in zip(patch_names, source_paths):
+            # Ensure destination exists (in case a prefix was supplied)
+            dirname = os.path.dirname(patch_name)
+            dest_path = os.path.join(self.path, dirname)
+            if dirname and not os.path.exists(dest_path):
+                os.makedirs(dest_path)
+
+            dest_path = os.path.join(self.path, patch_name)
+            if os.path.exists(dest_path):
+                # For simplicity, we regenerate all patches, however some will
+                # be the same, so perform a file compare so we keep accurate
+                # counts of which were truly updatd
+                if not filecmp.cmp(source_path, dest_path):
+                    updated.add(patch_name)
+                    shutil.move(source_path, dest_path)
+            else:
+                added.add(patch_name)
+                shutil.move(source_path, dest_path)
+
+        # Update series file
         with self._mutate_series_file() as entries:
             if parent_patch_name:
                 base = entries.index(parent_patch_name) + 1
@@ -543,9 +632,6 @@ class PatchRepo(git.Repo):
                     # Already exists, reorder patch by removing it from
                     # current location and inserting it into the new location.
                     entries.remove(patch_name)
-                    updated.add(patch_name)
-                else:
-                    added.add(patch_name)
 
                 entries.insert(base + idx, patch_name)
 
