@@ -1,3 +1,4 @@
+import contextlib
 import fnmatch
 import os
 import re
@@ -6,6 +7,16 @@ import re
 RE_PATCH_IDENTIFIER = re.compile('Ply-Patch: (.*)')
 
 
+@contextlib.contextmanager
+def usedir(path):
+    orig_path = os.getcwd()
+    os.chdir(path)
+    try:
+        yield
+    finally:
+        os.chdir(orig_path)
+
+
 def get_patch_annotation(commit_msg):
     """Return the Ply-Patch annotation if present in the commit msg.
 
